import copy
import traceback
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from supervisely.sly_logger import logger


class PackerUnpacker:
    SUPPORTED_TYPES = tuple([dict, pd.DataFrame])

    @staticmethod
    def validate_sizes(unpacked_data):
        if len(unpacked_data["data"]) != len(unpacked_data["classes"]):
            raise ValueError(
                "Sizes mismatch:\n"
                f'number of rows ({len(unpacked_data["data"])}) != number of columns ({len(unpacked_data["classes"])})'
            )

        for row in unpacked_data["data"]:
            if len(row) != len(unpacked_data["classes"]):
                raise ValueError(
                    "Sizes mismatch:\n"
                    f'{len(row)} != {len(unpacked_data["classes"])}\n'
                    f"{row}\n"
                    f'{unpacked_data["classes"]}'
                )

    @staticmethod
    def unpack_data(data, unpacker_cb, validate_sizes=True):
        unpacked_data = unpacker_cb(data)
        if validate_sizes:
            PackerUnpacker.validate_sizes(unpacked_data)
        return unpacked_data

    @staticmethod
    def pack_data(data, packer_cb):
        packed_data = packer_cb(data)
        return packed_data

    @staticmethod
    def dict_unpacker(data: dict):
        formatted_rows = []
        for origin_row in data["data"]:
            formatted_rows.append([{"value": element} for element in origin_row])

        unpacked_data = {"classes": data["columns"], "data": formatted_rows}

        return unpacked_data

    @staticmethod
    def pandas_unpacker(data: pd.DataFrame):
        data = data.where(pd.notnull(data), None)
        data = data.astype(object).replace(np.nan, "None")

        formatted_rows = []
        for origin_row in list(data.values):
            formatted_rows.append([{"value": element} for element in origin_row])

        unpacked_data = {"classes": data.columns.to_list(), "data": formatted_rows}
        return unpacked_data

    @staticmethod
    def dict_packer(data):
        unformatted_rows = []

        for origin_row in data["data"]:
            unformatted_rows.append([element["value"] for element in origin_row])

        packed_data = {
            "data": unformatted_rows,
            "columns": data["classes"],
        }
        return packed_data

    @staticmethod
    def pandas_packer(data):
        unformatted_rows = []

        for origin_row in data["data"]:
            unformatted_rows.append([element["value"] for element in origin_row])

        packed_data = pd.DataFrame(data=unformatted_rows, columns=data["classes"])
        return packed_data


DATATYPE_TO_PACKER = {pd.DataFrame: PackerUnpacker.pandas_packer, dict: PackerUnpacker.dict_packer}

DATATYPE_TO_UNPACKER = {
    pd.DataFrame: PackerUnpacker.pandas_unpacker,
    dict: PackerUnpacker.dict_unpacker,
}


class ConfusionMatrix(Widget):
    """ConfusionMatrix is a widget that display a given confusion matrix with color-coded visualization for better interpretation.
    It also shows row and column totals.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/charts-and-plots/confusionmatrix>`_
        (including screenshots and examples).


    :param data: Data of table in different formats (see usage example)
    :type data: Optional[Union[pd.DataFrame, Dict]]
    :param columns: List of column names
    :type columns: Optional[List[str]]
    :param x_label: Label for x axis
    :type x_label: Optional[str]
    :param y_label: Label for y axis
    :type y_label: Optional[str]
    :param widget_id: An unique identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import ConfusionMatrix

        # Option 1: Python dict
        confusion_matrix = ConfusionMatrix(
            data={
                "columns": ["class_1", "class_2", "class_3"],
                "data": [
                    ["1", "2", "3"],
                    ["4", "5", "6"],
                    ["7", "8", "9"],
                ],
            },
            x_label="Predicted Values",
            y_label="Actual Values",
        )

        # Option 2: Pandas DataFrame

        data = []
        for row in b:
            temp = [round(row * number, 1) for number in a]
            data.append(temp)

        a = [str(i) for i in a]
        b = [str(i) for i in b]

        data = pd.DataFrame(data=data, index=b, columns=a)
        confusion_matrix = ConfusionMatrix(data=data)

    """

    class Routes:
        CELL_CLICKED = "cell_clicked_cb"

    class ClickedDataPoint:
        """Represents data point of clicked cell in ConfusionMatrix.

        :param column_name: Name of column
        :type column_name: str
        :param column_index: Index of column
        :type column_index: int
        :param row_name: Name of row
        :type row_name: str
        :param row_index: Index of row
        :type row_index: int
        :param cell_value: Value of cell
        :type cell_value: Any
        """

        def __init__(
            self,
            column_name: str,
            column_index: int,
            row_name: str,
            row_index: int,
            cell_value: Any,
        ):
            self.column_index = column_index
            self.column_name = column_name
            self.row_index = row_index
            self.row_name = row_name
            self.cell_value = cell_value

    def __init__(
        self,
        data: Optional[Union[pd.DataFrame, Dict]] = None,
        columns: Optional[List[str]] = None,
        x_label: Optional[str] = "Predicted Values",
        y_label: Optional[str] = "Actual Values",
        widget_id: Optional[str] = None,
    ):
        self._supported_types = PackerUnpacker.SUPPORTED_TYPES

        self._parsed_data = None
        self._parsed_data_with_totals = {}
        self._data_type = None

        self._update_matrix_data(input_data=pd.DataFrame(data=data, columns=columns))

        self.x_label = x_label
        self.y_label = y_label

        self._loading = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - matrix_data: data of matrix
            - matrix_options: dictionary with options for matrix
                - selectable: if True, matrix will be selectable
                - horizontalLabel: label for horizontal axis
                - verticalLabel: label for vertical axis
            - loading: if True, loading animation will be shown

        :return: dictionary with widget data
        :rtype: Dict[str, Any]
        """

        return {
            "matrix_data": self._parsed_data_with_totals,
            "matrix_options": {
                "selectable": True,
                "horizontalLabel": self.x_label,
                "verticalLabel": self.y_label,
            },
            "loading": self._loading,
        }

    def get_json_state(self) -> Dict[str, Dict]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - selected_row: dictionary with empty dict

        :return: dictionary with widget state
        :rtype: Dict[str, Dict]
        """
        return {"selected_row": {}}

    def _update_matrix_data(self, input_data):
        if input_data is not None:
            self._parsed_data = copy.deepcopy(self.get_unpacked_data(input_data=input_data))
            self._data_type = type(input_data)
        else:
            self._parsed_data = {"classes": [], "data": []}
            self._data_type = dict
        self._calculate_totals()

    def _calculate_totals(self):
        matrix_data = []
        for origin_row in self._parsed_data["data"]:
            matrix_data.append([element["value"] for element in origin_row])

        totals_by_rows = np.asarray(["-" for _ in matrix_data]).reshape(-1, 1)
        totals_by_columns = np.asarray([["-" for _ in matrix_data]])

        try:
            matrix_data = np.matrix(matrix_data).astype(float)
            totals_by_rows = np.sum(matrix_data, axis=1).round(2)
            totals_by_columns = np.sum(matrix_data, axis=0).round(2)

            self._parsed_data_with_totals.update(self._calculate_max_values(matrix_data))

        except Exception as ex:
            logger.warning(f"Cannot calculate totals for matrix ({self.__class__.__name__}): {ex}")

        totals_by_columns = np.hstack([totals_by_columns, [[None]]])

        matrix_data = np.hstack([matrix_data, totals_by_rows])
        matrix_data = np.vstack([matrix_data, totals_by_columns])

        self._parsed_data_with_totals.update(
            self.get_unpacked_data(
                input_data={"columns": self._parsed_data["classes"], "data": matrix_data.tolist()},
                validate_sizes=False,
            )
        )

    def _get_packed_data(self, input_data, data_type):
        return PackerUnpacker.pack_data(data=input_data, packer_cb=DATATYPE_TO_PACKER[data_type])

    def get_unpacked_data(
        self, input_data, validate_sizes: Optional[bool] = True
    ) -> Dict[str, Any]:
        """Returns dictionary with unpacked data.

        :param input_data: Data of table in different formats (see usage example)
        :type input_data: Optional[Union[pd.DataFrame, Dict]]
        :param validate_sizes: If True, sizes of data will be validated, defaults to True
        :type validate_sizes: Optional[bool], optional
        :raises TypeError: If input data type is not supported
        :return: dictionary with unpacked data
        :rtype: Dict[str, Any]
        """
        input_data_type = type(input_data)

        if input_data_type not in self._supported_types:
            raise TypeError(
                f"Cannot parse input data, please use one of supported datatypes: {self._supported_types}\n"
                """
                            1. Pandas Dataframe \n
                            2. Python dict with structure {
                                        'columns': ['col_name_1', 'col_name_2', ...],
                                        'data': [
                                                            ['row_1_column_1', 'row_1_column_2', ...],
                                                            ['row_2_column_1', 'row_2_column_2', ...],
                                                            ...
                                                          ]
                                      }
                            """
            )

        return PackerUnpacker.unpack_data(
            data=input_data,
            unpacker_cb=DATATYPE_TO_UNPACKER[input_data_type],
            validate_sizes=validate_sizes,
        )

    def to_json(self) -> Dict[str, Any]:
        """Returns dictionary with widget data in JSON format.

        :return: dictionary with widget data in JSON format
        :rtype: Dict[str, Any]
        """
        return self._get_packed_data(self._parsed_data, dict)

    def to_pandas(self) -> pd.DataFrame:
        """Returns pandas DataFrame with widget data.

        :return: pandas DataFrame with widget data
        :rtype: pd.DataFrame
        """
        return self._get_packed_data(self._parsed_data, pd.DataFrame)

    def read_json(self, value: Dict[str, Any]) -> None:
        """Sets widget data from JSON.

        :param value: dictionary with widget data in JSON format
        :type value: Dict[str, Any]
        """
        self._update_matrix_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data

    def read_pandas(self, value: pd.DataFrame) -> None:
        """Sets widget data from pandas DataFrame.

        :param value: pandas DataFrame with widget data
        :type value: pd.DataFrame
        """
        self._update_matrix_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data

    def get_selected_cell(self, state: Dict[str, Dict]) -> Dict[str, Any]:
        """Returns dictionary with selected cell data.

        Dictionary contains the following fields:
            - row_index: index of row
            - col_index: index of column
            - row_data: data of row
            - cell_data: data of cell

        :param state: dictionary with widget state
        :type state: Dict[str, Dict]
        :return: dictionary with selected cell data
        :rtype: Dict[str, Any]
        """
        row_index = state[self.widget_id]["selected_row"].get("row")
        col_index = state[self.widget_id]["selected_row"].get("col")

        row_data = None
        cell_data = None

        if row_index is not None and col_index is not None:
            row_data = [element["value"] for element in self._parsed_data["data"][row_index]]
            cell_data = {
                "row_name": self._parsed_data["classes"][row_index],
                "col_name": self._parsed_data["classes"][col_index],
                "value": row_data[col_index],
            }

        return {
            "row_index": row_index,
            "col_index": col_index,
            "row_data": row_data,
            "cell_data": cell_data,
        }

    def _calculate_max_values(self, matrix_data):
        return {
            "diagonalMax": float(max(np.diagonal(matrix_data))),
            "maxValue": float(matrix_data.max()),
        }

    @property
    def loading(self) -> bool:
        """Returns True if loading animation is shown, False otherwise.

        :return: True if loading animation is shown, False otherwise
        :rtype: bool
        """
        return self._loading

    @loading.setter
    def loading(self, value: bool) -> None:
        """Sets loading animation.

        :param value: True if loading animation is shown, False otherwise
        :type value: bool
        """
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading

    def click(self, func: Callable[[ClickedDataPoint], Any]) -> Callable[[], None]:
        """Decorator for function that will be called when cell is clicked.

        :param func: function that will be called when cell is clicked
        :type func: Callable[[ClickedDataPoint], Any]
        :return: decorated function
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(ConfusionMatrix.Routes.CELL_CLICKED)
        server = self._sly_app.get_server()

        @server.post(route_path)
        def _click():
            try:
                value_dict = self.get_selected_cell(StateJson())
                if value_dict is None:
                    return
                datapoint = ConfusionMatrix.ClickedDataPoint(
                    column_name=value_dict["cell_data"]["col_name"],
                    column_index=value_dict["col_index"],
                    row_name=value_dict["cell_data"]["row_name"],
                    row_index=value_dict["row_index"],
                    cell_value=value_dict["cell_data"]["value"],
                )
                func(datapoint)
            except Exception as e:
                logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
                raise e

        return _click



import supervisely.nn.checkpoints.yolov5 as yolov5
import supervisely.nn.checkpoints.yolov5_v2 as yolov5_v2
import supervisely.nn.checkpoints.yolov8 as yolov8
import supervisely.nn.checkpoints.detectron2 as detectron2
import supervisely.nn.checkpoints.mmdetection as mmdetection
import supervisely.nn.checkpoints.mmdetection3 as mmdetection3
import supervisely.nn.checkpoints.mmsegmentation as mmsegmentation
import supervisely.nn.checkpoints.mmclassification as mmclassification
import supervisely.nn.checkpoints.ritm as ritm
import supervisely.nn.checkpoints.unet as unet
import supervisely.nn.checkpoints.hrda as hrda



from os.path import basename, join
from typing import List

from supervisely.api.api import Api
from supervisely.nn.checkpoints.checkpoint import CheckpointInfo


def get_list(api: Api, team_id: int) -> List[CheckpointInfo]:
    checkpoints = []
    weights_dir_name = "weights"
    training_app_directory = "/yolov8_train/"
    if not api.file.dir_exists(team_id, training_app_directory):
        return []
    object_detection_projects_dir = join(training_app_directory, "object detection")
    instance_segmentation_projects_dir = join(training_app_directory, "instance segmentation")
    pose_estimation_projects_dir = join(training_app_directory, "pose estimation")
    paths_to_projects = [
        object_detection_projects_dir,
        instance_segmentation_projects_dir,
        pose_estimation_projects_dir,
    ]
    for path_to_projects in paths_to_projects:
        if not api.file.dir_exists(team_id, path_to_projects):
            continue
        task_type = basename(path_to_projects).strip()
        project_files_infos = api.file.list(team_id, path_to_projects, recursive=False)
        for project_file_info in project_files_infos:
            project_name = project_file_info["name"]
            task_files_infos = api.file.list(team_id, project_file_info["path"], recursive=False)
            for task_file_info in task_files_infos:
                if task_file_info["name"] == "images":
                    continue
                task_id = task_file_info["name"]
                session_link = f"{api.server_address}/apps/sessions/{task_id}"
                paths_to_checkpoints = join(task_file_info["path"], weights_dir_name)
                checkpoints_infos = api.file.list(team_id, paths_to_checkpoints, recursive=False)
                if len(checkpoints_infos) == 0:
                    continue
                checkpoint_info = CheckpointInfo(
                    app_name="Train YOLOv8",
                    session_id=task_id,
                    session_path=task_file_info["path"],
                    session_link=session_link,
                    task_type=task_type,
                    training_project_name=project_name,
                    checkpoints=checkpoints_infos,
                )
                checkpoints.append(checkpoint_info)
    return checkpoints



import os
from dotenv import load_dotenv

import supervisely as sly


# load_dotenv("local.env")
load_dotenv(os.path.expanduser("~/supervisely.env"))

api = sly.Api()

# team_id = sly.env.team_id()
team_id = 435
PATH = "/sly-app-data"


files_list = api.file.list(team_id, PATH)

print('Total {} files in "{}" directory.'.format(len(files_list), PATH))
print(f"First file in list: {files_list[0]}")
# Output: 
# Total 29 files in "/sly-app-data" directory.
# First file in list: {'id': 772315, 'isDir': True, 'meta': {'size': 8605}, 'path': '/sly-app-data/ma...}


files_list_2 = api.file.list2(team_id, PATH)

print('Total {} files in "{}" directory.'.format(len(files_list_2), PATH))
print(f"First FileInfo object in list: {files_list_2[0]}")
# Output: 
# Total 29 files in "/sly-app-data" directory.
# First FileInfo object in list: FileInfo(team_id=435, id=18421, user_id=330, name='507_002.tar.gz', hash='+0nrNoDjBxxJA...


listdir = api.file.listdir(team_id, PATH)

print(listdir)
# Output: [
#     '/sly-app-data/object-tags-editor-files',
#     '/sly-app-data/mark-segments-on-synced-videos-2-files'
# ]


listdir_s = api.file.listdir(team_id, f"{PATH}/")

print(listdir_s)
# before update SDK
# Output: [] # return [] if path ends with "/"

# after update SDK pr#382 
# Output: [
#     '/sly-app-data/object-tags-editor-files',
#     '/sly-app-data/mark-segments-on-synced-videos-2-files'
# ]


###############################################################################
# https://github.com/supervisely/supervisely/pull/390
# added new parameter recursive in list, list2, listdir methods
# fixed path separator in listdir method


list_not_recursively = api.file.list(team_id, PATH, recursive=False) # default recursive=True

print(list_not_recursively)
# Output: [
#     {'id': 772315, 'isDir': True, 'meta': {'size': 8605}, 'path': '/sly-app-data/ma...},
#     {'id': 847993, 'isDir': True, 'meta': {'size': 277}, 'path': '/sly-app-data/ob...}
# ]


list2_not_recursively = api.file.list2(team_id, PATH, recursive=False) # default recursive=True

print(list2_not_recursively)
# Output: [
#     FileInfo(team_id=435, id=18421, user_id=330, name='50t_video_002.tar.gz', hash='+0nrNoDjBxxJA...
#     FileInfo(team_id=435, id=18431, user_id=330, name='50t_video_002.tar.gz', hash='+0nrNoDjBxxJA...
# ]


listdir_recursively = api.file.listdir(team_id, PATH, recursive=True)

print('{} paths has been collected from "{}" directory.'.format(len(listdir_recursively), PATH))
print(f"First path in list: {listdir_recursively[0]}")
# Output: 
# 29 paths has been collected from "/sly-app-data" directory.
# "First path in list: '/sly-app-data/mark-segments-on-synced-videos-2-files/project-15918/dataset-54074/Info.json'


listdir_recursively_s = api.file.listdir(team_id, f"{PATH}/", recursive=True)

print('{} paths has been collected from "{}" directory.'.format(len(listdir_recursively_s), PATH))
print(f"First path in list: {listdir_recursively_s[0]}")
# Output: 
# 29 paths has been collected from "/sly-app-data" directory.
# "First path in list: '/sly-app-data/mark-segments-on-synced-videos-2-files/project-15918/dataset-54074/Info.json'



import os
from typing import List, Union

import magic
import nrrd
import numpy as np

from supervisely import Annotation, ProjectMeta, logger
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.convert.image.medical2d import medical2d_helper as helper
from supervisely.io.fs import clean_dir, get_file_ext, mkdir
from supervisely.volume.volume import is_nifti_file


# @TODO: add group tags?
class Medical2DImageConverter(ImageConverter):

    def __init__(self, input_data: str, labeling_interface: str) -> None:
        self._input_data: str = input_data
        self._items: List[ImageConverter.Item] = []
        self._meta: ProjectMeta = None
        self._labeling_interface = labeling_interface
        self._filtered = None

    def __str__(self):
        return AvailableImageConverters.MEDICAL2D

    def validate_labeling_interface(self) -> bool:
        """Only medical labeling interface can be used for medical images."""
        return self._labeling_interface == "medical_imaging_single"

    def validate_format(self) -> bool:
        logger.debug(f"Validating format: {self.__str__()}")

        converted_dir_name = "converted"
        converted_dir = os.path.join(self._input_data, converted_dir_name)
        mkdir(converted_dir, remove_content_if_exists=True)

        nrrd = {}
        for root, _, files in os.walk(self._input_data):
            if converted_dir_name in root:
                continue
            for file in files:
                path = os.path.join(root, file)
                ext = get_file_ext(path).lower()
                mime = magic.from_file(path, mime=True)
                if mime == "application/dicom":
                    if helper.is_dicom_file(path):  # is dicom
                        paths, names = helper.convert_dcm_to_nrrd(path, converted_dir)
                        for path, name in zip(paths, names):
                            nrrd[name] = path
                elif ext == ".nrrd":
                    if helper.check_nrrd(path):  # is nrrd
                        paths, names = helper.slice_nrrd_file(path, converted_dir)
                        for path, name in zip(paths, names):
                            nrrd[name] = path
                elif mime == "application/gzip" or mime == "application/octet-stream":
                    if is_nifti_file(path):  # is nifti
                        paths, names = helper.slice_nifti_file(path, converted_dir)
                        for path, name in zip(paths, names):
                            nrrd[name] = path

        self._items = []
        for name, path in nrrd.items():
            item = self.Item(item_path=path)
            self._items.append(item)

        if not self._items:
            clean_dir(converted_dir)

        return self.items_count > 0

    def _get_image_channels(self, file_path: str) -> List[np.ndarray]:
        file_ext = get_file_ext(file_path).lower()
        logger.debug(f"Working with file {file_path} with extension {file_ext}.")

        if file_ext == ".nrrd":
            logger.debug(f"Found nrrd file: {file_path}.")
            image, _ = nrrd.read(file_path)
        elif file_ext == ".dcm":
            pass
        return [image[:, :, i] for i in range(image.shape[2])]

    def to_supervisely(self, item, meta, *args) -> Union[Annotation, None]:
        return None



from typing import Any, List, Optional
from supervisely.app.widgets import Widget


class OptionComponent:
    def __init__(
        self,
        component_name: str,
        default_value: Optional[Any] = None,
        sidebar_component: Optional[str] = None,
        options: Optional[dict] = {},
    ):
        self.component_name = component_name
        self.default_value = default_value
        self.sidebar_component = sidebar_component
        self.options = options

    def to_json(self):
        return {
            "component": self.component_name,
            "value": self.default_value,
            "sidebarComponent": self.sidebar_component,
            "options": self.options,
        }


class HtmlOptionComponent(OptionComponent):
    def __init__(
        self,
        html: str,
        sidebar_component: Optional[OptionComponent] = None,
        sidebar_width: Optional[int] = None,
    ):
        options = {}
        sidebar_component_name = None
        if sidebar_component is not None:
            sidebar_component_name = sidebar_component.component_name
            sidebar_template = sidebar_component.options["template"]
            options["sidebarTemplate"] = f"<div>{sidebar_template}</div>"
            options["sidebarWidth"] = sidebar_width
        options["template"] = f"<div>{html}</div>"
        super().__init__(
            component_name="SlyFlowOptionRenderer",
            options=options,
            sidebar_component=sidebar_component_name
        )


class WidgetOptionComponent(HtmlOptionComponent):
    def __init__(self, widget: Widget, sidebar_component=None, sidebar_width: Optional[int] = None):
        super().__init__(
            widget.to_html(), sidebar_component=sidebar_component, sidebar_width=sidebar_width
        )


class ButtonOptionComponent(OptionComponent):
    """A button that opens the sidebar when clicked. The label of the button is determined by the name of the option."""

    def __init__(
        self,
        sidebar_component: Optional[OptionComponent] = None,
        sidebar_width: Optional[int] = None,
    ):
        sidebar_component_name = None
        options = {}
        if sidebar_component is not None:
            sidebar_component_name = sidebar_component.component_name
            options = sidebar_component.options
            options["sidebarWidth"] = sidebar_width
        super().__init__(
            component_name="ButtonOption", sidebar_component=sidebar_component_name, options=options
        )


class CheckboxOptionComponent(OptionComponent):
    """A checkbox for setting boolean values."""

    def __init__(self, default_value: bool = False):
        super().__init__(component_name="CheckboxOption", default_value=default_value)


class InputOptionComponent(OptionComponent):
    """A simple text field. The option name will be displayed as placeholder."""

    def __init__(self, default_value: str = ""):
        super().__init__(component_name="InputOption", default_value=default_value)


class IntegerOptionComponent(OptionComponent):
    """A numeric up/down field for integer values."""

    def __init__(
        self,
        min: Optional[int] = None,
        max: Optional[int] = None,
        default_value: Optional[int] = 0,
    ):
        options = {}
        if min is not None:
            options["min"] = min
        if max is not None:
            options["max"] = max
        super().__init__(
            component_name="IntegerOption",
            default_value=default_value,
            options=options,
        )


class NumberOptionComponent(OptionComponent):
    """A numeric up/down field for numeric values."""

    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        default_value: Optional[float] = 0,
    ):
        options = {}
        if min is not None:
            options["min"] = min
        if max is not None:
            options["max"] = max
        super().__init__(
            component_name="NumberOption",
            default_value=default_value,
            options=options,
        )


class SelectOptionComponent(OptionComponent):
    """A dropdown select which allows the user to choose one of predefined values."""

    class Item:
        def __init__(self, value: str, label: Optional[str] = None):
            self.value = value
            self.label = value if label is None else label

        def to_json(self):
            return {"text": self.label, "value": self.value}

    def __init__(self, items: List[Item], default_value: Optional[str] = None):
        options = {"items": [item.to_json() for item in items]}
        super().__init__(
            component_name="SelectOption",
            default_value=default_value,
            options=options,
        )


class SliderOptionComponent(OptionComponent):
    """A slider for choosing a value in a range. The user can also click on the slider to set a specific value."""

    def __init__(
        self,
        min: float,
        max: float,
        default_value: Optional[float] = None,
    ):
        options = {
            "min": min,
            "max": max,
        }
        default_value = min if default_value is None else default_value
        super().__init__(
            component_name="SliderOption",
            default_value=default_value,
            options=options,
        )


class TextOptionComponent(OptionComponent):
    """Displays arbitrary strings"""

    def __init__(self, text: str):
        super().__init__(component_name="TextOption", default_value=text)


class SidebarNodeInfoOptionComponent(OptionComponent):
    def __init__(self, sidebar_template: str, sidebar_width: Optional[int] = None):
        options = {
            "sidebarTemplate": f"<div>{sidebar_template}</div>", 
            "sidebarWidth": sidebar_width
        }
        super().__init__(
            component_name="SlyFlowOptionRenderer",
            sidebar_component="SlyFlowOptionRenderer",
            options=options,
        )



# coding: utf-8

import os
import os.path
import json
from copy import deepcopy

from supervisely import logger
from supervisely.io import fs as sly_fs
from supervisely.nn.legacy.config import update_recursively
from supervisely.nn.legacy.dataset import samples_by_tags
from supervisely.nn.legacy.hosted import class_indexing
from supervisely.nn.legacy.hosted.constants import SETTINGS, INPUT_SIZE, HEIGHT, WIDTH
from supervisely.task.progress import report_checkpoint_saved
from supervisely.project.project import read_single_project
from supervisely.geometry.rectangle import Rectangle
from supervisely.task.paths import TaskPaths
from supervisely.io.json import dump_json_file


BATCH_SIZE = 'batch_size'
LR = 'lr'
DATASET_TAGS = 'dataset_tags'
EPOCHS = 'epochs'
LOSS = 'loss'
TRAIN = 'train'
VAL = 'val'
WEIGHTS_INIT_TYPE = 'weights_init_type'


class TrainCheckpoints:
    def __init__(self, base_out_dir):
        self._base_out_dir = base_out_dir
        # Checkpoint index does not correspond to epoch. Depending on training config, checkpoints may be saved more
        # frequently than once per epoch (or less frequently than once per epoch).
        self._idx = -1
        self._prepare_next_dir()
        self._last_saved_dir = None

    def _prepare_next_dir(self):
        self._idx += 1
        self._subdir = '{:08}'.format(self._idx)
        self._odir = os.path.join(self._base_out_dir, self._subdir)
        sly_fs.mkdir(self._odir)

    def saved(self, is_best, sizeb, optional_data=None):
        report_best = is_best or (self._idx == 0)
        report_checkpoint_saved(checkpoint_idx=self._idx, subdir=self._subdir, sizeb=sizeb, best_now=report_best,
                                optional_data=optional_data)
        self._last_saved_dir = self._odir
        self._prepare_next_dir()

    def get_dir_to_write(self):
        return self._odir

    def get_last_ckpt_dir(self):
        return self._last_saved_dir


class SuperviselyModelTrainer:
    """
    Base class for train neural networks with Supervisely.

    It is highly recommended that your train classes subclass this class.

    Args:
        default_config: Dict object containing default training config.
    """
    def __init__(self, default_config):
        logger.info('Will init all required to train.')

        with open(TaskPaths.TASK_CONFIG_PATH) as fin:
            self._task_config = json.load(fin)

        self.checkpoints_saver = TrainCheckpoints(TaskPaths.RESULTS_DIR)
        self.project = read_single_project(TaskPaths.DATA_DIR)
        logger.info('Project structure has been read. Samples: {}.'.format(self.project.total_items))

        self._default_config = deepcopy(default_config)
        self._determine_config()
        self._determine_model_classes()
        self._determine_out_config()
        self._construct_samples_dct()
        self._construct_data_loaders()
        self._construct_and_fill_model()
        self._construct_loss()

        self.epoch_flt = 0

    @property
    def class_title_to_idx_key(self):
        return 'class_title_to_idx'

    @property
    def train_classes_key(self):
        return 'classes'

    def get_start_class_id(self):
        # Returns the first integer id to use when assigning integer ids to class names.
        # The usual setting for segmentation network, where background is often a special class with id 0.
        return 1

    def _validate_train_cfg(self, config):
        pass

    def _determine_config(self):
        """
        Determines train config by updating default config with user config and then validates them with given
        json schema.
        """
        input_config = self._task_config
        logger.info('Input config', extra={'config': input_config})
        config = deepcopy(self._default_config)
        update_recursively(config, input_config)
        logger.info('Full config', extra={'config': config})

        self._validate_train_cfg(config)
        self.config = config
        input_size_dict = self.config[INPUT_SIZE]
        self._input_size = (input_size_dict[HEIGHT], input_size_dict[WIDTH])

    def _determine_model_classes_segmentation(self, bkg_input_idx):
        self.out_classes, self.class_title_to_idx = class_indexing.create_segmentation_classes(
            in_project_classes=self.project.meta.obj_classes,
            special_classes_config=self.config.get('special_classes', {}),
            bkg_input_idx=bkg_input_idx,
            weights_init_type=self.config[WEIGHTS_INIT_TYPE],
            model_config_fpath=TaskPaths.MODEL_CONFIG_PATH,
            class_to_idx_config_key=self.class_title_to_idx_key,
            start_class_id=self.get_start_class_id())

    def _determine_model_classes_detection(self):
        in_project_classes = self.project.meta.obj_classes
        self.out_classes = class_indexing.make_out_classes(in_project_classes, geometry_type=Rectangle)
        logger.info('Determined model out classes', extra={'classes': list(self.out_classes)})
        in_project_class_to_idx = class_indexing.make_new_class_to_idx_map(
            in_project_classes, start_class_id=self.get_start_class_id())
        self.class_title_to_idx = class_indexing.infer_training_class_to_idx_map(
            self.config[WEIGHTS_INIT_TYPE],
            in_project_class_to_idx,
            TaskPaths.MODEL_CONFIG_PATH,
            class_to_idx_config_key=self.class_title_to_idx_key)
        logger.info('Determined class mapping.', extra={'class_mapping': self.class_title_to_idx})

    def _determine_model_classes(self):
        # Use _determine_model_classes_segmentation() or _determine_model_classes_detection() here depending on the
        # model needs.
        raise NotImplementedError()

    def _determine_out_config(self):
        self.out_config = {
            SETTINGS: self.config,
            self.train_classes_key: self.out_classes.to_json(),
            self.class_title_to_idx_key: self.class_title_to_idx,
        }

    def _construct_and_fill_model(self):
        raise NotImplementedError()

    def _construct_loss(self):
        # Useful for Tensorflow based models.
        raise NotImplementedError()

    def _construct_samples_dct(self):
        logger.info('Will collect samples (image/annotation pairs).')
        self.name_to_tag = self.config[DATASET_TAGS]
        self._deprecated_samples_by_tag = samples_by_tags(required_tags=list(self.name_to_tag.values()), project=self.project)
        self._samples_by_data_purpose = {purpose: self._deprecated_samples_by_tag[tag] for purpose, tag in self.config[DATASET_TAGS].items()}

    def _construct_data_loaders(self):
        # Pipeline-specific code to set up data loading should go here.
        raise NotImplementedError()

    def _dump_model_weights(self, out_dir):
        raise NotImplementedError

    def _save_model_snapshot(self, is_best, opt_data):
        out_dir = self.checkpoints_saver.get_dir_to_write()
        dump_json_file(self.out_config, os.path.join(out_dir, TaskPaths.MODEL_CONFIG_NAME))
        self._dump_model_weights(out_dir)
        size_bytes = sly_fs.get_directory_size(out_dir)
        self.checkpoints_saver.saved(is_best, size_bytes, opt_data)

    def train(self):
        """
        Start train process.
        """
        raise NotImplementedError()



from os.path import join
from typing import List

from supervisely.api.api import Api
from supervisely.nn.checkpoints.checkpoint import CheckpointInfo


def get_list(api: Api, team_id: int) -> List[CheckpointInfo]:
    checkpoints = []
    weights_dir_name = "weights"
    training_app_directory = "/yolov5_train/"
    task_type = "object detection"
    if not api.file.dir_exists(team_id, training_app_directory):
        return []
    project_files_infos = api.file.list(team_id, training_app_directory, recursive=False)
    for project_file_info in project_files_infos:
        project_name = project_file_info["name"]
        task_files_infos = api.file.list(team_id, project_file_info["path"], recursive=False)
        for task_file_info in task_files_infos:
            if task_file_info["name"] == "images":
                continue
            task_id = task_file_info["name"]
            session_link = f"{api.server_address}/apps/sessions/{task_id}"
            paths_to_checkpoints = join(task_file_info["path"], weights_dir_name)
            checkpoints_infos = api.file.list(team_id, paths_to_checkpoints, recursive=False)
            if len(checkpoints_infos) == 0:
                continue
            checkpoint_info = CheckpointInfo(
                app_name="Train YOLOv5",
                session_id=task_id,
                session_path=task_file_info["path"],
                session_link=session_link,
                task_type=task_type,
                training_project_name=project_name,
                checkpoints=checkpoints_infos,
            )
            checkpoints.append(checkpoint_info)
    return checkpoints



import csv
import os
import os.path
from typing import Tuple

import pandas as pd

from supervisely import ProjectMeta, TagMeta, TagMetaCollection, TagValueType, logger
from supervisely.io.fs import silent_remove

DEFAULT_DELIMITER = ";"
INCORRECT_COLUMN_DELIMITER = ","
TAGS_DELIMITER = ","
possible_image_url_col_names = ["url"]
possible_image_path_col_names = ["path"]
possible_tag_col_names = ["tag"]

# ------------------------------------------- Convert ------------------------------------------- #


def rename_in_json(ann_json, renamed_classes=None, renamed_tags=None):
    if renamed_classes:
        for obj in ann_json["objects"]:
            obj["classTitle"] = renamed_classes.get(obj["classTitle"], obj["classTitle"])
    if renamed_tags:
        for tag in ann_json["tags"]:
            tag["name"] = renamed_tags.get(tag["name"], tag["name"])
    return ann_json


def convert_txt_to_csv(txt_file_path: str, csv_file_path: str, delimiter: str = r"\s{2,}|\t+|;"):
    data = pd.read_csv(txt_file_path, delimiter=delimiter, engine="python")
    data.to_csv(csv_file_path, index=False, sep=DEFAULT_DELIMITER)


def convert_tsv_to_csv(tsv_file_path: str, csv_file_path: str, delimiter: str = DEFAULT_DELIMITER):
    data = pd.read_csv(tsv_file_path, delimiter="\t")
    data.to_csv(csv_file_path, index=False, sep=delimiter)


# ------------------------------------------- Validate ------------------------------------------- #


def flat_tag_list(total_tags):
    tags = []
    for tag in total_tags:
        if tag is None or tag == "":
            continue
        else:
            if TAGS_DELIMITER in tag:
                tag = tag.split(TAGS_DELIMITER)
            tags.append(tag)

    total_tags = []
    for sublist in tags:
        if type(sublist) == str:
            total_tags.append(sublist)
        else:
            for tag in sublist:
                if tag != "":
                    total_tags.append(tag)

    total_tags = list(set(total_tags))
    return total_tags


def create_project_meta_from_csv_tags(total_tags):
    tag_metas = []
    for tag_name in total_tags:
        tag_meta = TagMeta(tag_name, TagValueType.NONE)
        tag_metas.append(tag_meta)
    tag_meta_col = TagMetaCollection(tag_metas)
    project_meta = ProjectMeta(tag_metas=tag_meta_col)
    return project_meta


def check_column_names(col_names_validate):
    if not any(
        image_url_name in possible_image_url_col_names for image_url_name in col_names_validate
    ) and not any(
        image_path_name in possible_image_path_col_names for image_path_name in col_names_validate
    ):
        raise ValueError(
            f"Invalid column name for image: '{col_names_validate}' in the first line. "
            f"Please use: '{ possible_image_url_col_names[0]}' for URL "
            f"or '{ possible_image_path_col_names[0]}' for path in Team Files"
        )
    if len(col_names_validate) == 2:
        if not any(tag_name in possible_tag_col_names for tag_name in col_names_validate):
            raise ValueError(
                f"Invalid column name for tag: {col_names_validate[1]}. Please use:{ possible_tag_col_names[0]}"
            )


def validate_column_names(first_csv_row: dict):
    col_names_validate = [key.lower() for key in first_csv_row.keys()]
    col_names = [key for key in first_csv_row.keys()]
    check_column_names(col_names_validate)
    image_col_name = None
    tag_col_name = None
    for name in col_names:
        if name == "url":
            image_col_name = name
        if name == "path":
            image_col_name = name
        if name == "tag":
            tag_col_name = name

    return image_col_name, tag_col_name


def first_line_is_url(first_line: str):
    first_line = first_line.strip().lstrip('"').lstrip("'")
    markers = ["https://", "http://", "s3://", "azure://", "google://"]
    for marker in markers:
        if first_line.startswith(marker):
            return True
    return False


def first_line_is_path(first_line):
    first_line = first_line.strip().lstrip('"').lstrip("'")
    if first_line.startswith("/") or first_line.startswith("agent://"):
        return True
    return False


def handle_csv_header(csv_path):
    content = []
    updated = False
    with open(csv_path, "r") as file:
        # read the file, correct it and write it back
        content = file.readlines()
        if len(content) == 0:
            raise ValueError(f"File '{csv_path}' is empty")
        first_line = content[0]
        has_tags = False
        new_header = None
        for line in content:
            if DEFAULT_DELIMITER in line:
                has_tags = True
                break
        if first_line_is_url(first_line):
            logger.debug("Headless csv detected. Adding 'url' column name to the first line.")
            new_header = "url" + DEFAULT_DELIMITER + "tag\n" if has_tags else "url\n"
        elif first_line_is_path(first_line):
            logger.debug("Headless csv detected. Adding 'path' column name to the first line.")
            new_header = "path" + DEFAULT_DELIMITER + "tag\n" if has_tags else "path\n"
        if new_header is None:
            first_line = content[1]
            if first_line_is_url(first_line):
                new_header = "url" + DEFAULT_DELIMITER + "tag\n" if has_tags else "url\n"
            elif first_line_is_path(first_line):
                new_header = "path" + DEFAULT_DELIMITER + "tag\n" if has_tags else "path\n"
            if new_header is not None:
                content[0] = new_header
                updated = True
            else:
                raise ValueError(
                    f"Invalid header in csv file. "
                    f"Please use: 'url' for URLs or 'path' for paths in Team Files"
                )
        else:
            content.insert(0, new_header)
            updated = True

    if updated:
        silent_remove(csv_path)
        with open(csv_path, "w") as file:
            file.writelines(content)


def validate_and_collect_items(csv_path) -> Tuple[dict, ProjectMeta, list]:
    csv_table = {"columns": [], "data": []}

    if os.path.getsize(csv_path) == 0:
        raise ValueError(f"File '{csv_path}' is empty")
    handle_csv_header(csv_path)
    with open(csv_path, "r") as images_csv:
        try:
            reader = csv.DictReader(images_csv, delimiter=DEFAULT_DELIMITER)
            if len(reader.fieldnames) > 0 and INCORRECT_COLUMN_DELIMITER in reader.fieldnames[0]:
                raise ValueError(
                    f"Invalid column delimiter: '{ INCORRECT_COLUMN_DELIMITER}'. "
                    f"Please, use: '{ DEFAULT_DELIMITER}' - for column delimiter and '{ TAGS_DELIMITER}' - for tags delimiter"
                )
            reader = [row for row in reader]
        except Exception as e:
            raise Exception(f"Can't read csv file. {e}")
        stripped_reader = []
        if len(reader) == 0:
            raise ValueError(f"File '{csv_path}' is empty")
        logger.debug(f"Total rows in csv file: {len(reader)}")
        for row in reader:
            stripped_row = {
                k: v.strip() for k, v in row.items() if k is not None and type(v) == str
            }
            stripped_reader.append(stripped_row)
        csv_reader = stripped_reader
        row_to_validate = {}
        for row in stripped_reader:
            if len(row) > len(row_to_validate) and len(row) < 3:
                row_to_validate = row
        if len(row_to_validate) == 0:
            row_to_validate = stripped_reader[0]
        image_col_name, tag_col_name = validate_column_names(row_to_validate)

        if tag_col_name is not None:
            csv_table["columns"] = ["row", image_col_name, tag_col_name]
            for idx, row in enumerate(stripped_reader):
                csv_table["data"].append([idx + 1, row[image_col_name], row.get(tag_col_name, "")])
            total_tags = flat_tag_list(
                list(set([row.get(tag_col_name) for row in stripped_reader]))
            )
            project_meta = create_project_meta_from_csv_tags(total_tags)
        else:
            csv_table["columns"] = ["row", image_col_name]
            for idx, row in enumerate(stripped_reader):
                csv_table["data"].append([idx + 1, row[image_col_name]])
            total_tags = 0
            project_meta = ProjectMeta()

    return project_meta, csv_reader



from supervisely.convert.base_converter import AvailableVideoConverters, BaseConverter


class DavisConverter(BaseConverter):
    def __init__(self, input_data, labeling_interface: str):
        super().__init__(input_data, labeling_interface)

    def __str__(self):
        return AvailableVideoConverters.DAVIS

    def get_meta(self):
        return super().get_meta()

    def get_items(self):
        return super().get_items()

    def to_supervisely(self, image_path: str, ann_path: str):
        raise NotImplementedError()



import os
from typing import Callable, List, NamedTuple, Optional, Union

from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

import supervisely.volume.nrrd_encoder as nrrd_encoder
from supervisely import logger, volume
from supervisely._utils import batched, generate_free_name
from supervisely.api.module_api import ApiField, RemoveableBulkModuleApi
from supervisely.api.volume.volume_annotation_api import VolumeAnnotationAPI
from supervisely.api.volume.volume_figure_api import VolumeFigureApi
from supervisely.api.volume.volume_object_api import VolumeObjectApi

# from supervisely.api.volume.video_frame_api import VolumeFrameAPI
from supervisely.api.volume.volume_tag_api import VolumeTagApi
from supervisely.imaging.image import read_bytes
from supervisely.io.fs import (
    ensure_base_path,
    get_bytes_hash,
    get_file_ext,
    get_file_name,
    get_file_name_with_ext,
)
from supervisely.task.progress import Progress, tqdm_sly
from supervisely.volume.volume import inspect_dicom_series, inspect_nrrd_series
from supervisely.volume_annotation.plane import Plane

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class VolumeInfo(NamedTuple):
    """
    Object with :class:`Volume<supervisely.volume.volume>` parameters from Supervisely.

    :Example:

     .. code-block:: python

        VolumeInfo(
            id=19581134,
            name='CTChest.nrrd',
            link=None,
            hash='+0K2oFHpqA5dwRKQlhCiiE2qwLNP76Xk0kvhXEh52cs=',
            mime=None,
            ext=None,
            sizeb=46073411,
            created_at='2023-03-29T12:30:37.078Z',
            updated_at='2023-03-29T12:30:37.078Z',
            meta={
                'ACS': 'RAS',
                'intensity': {'max': 3071, 'min': -3024},
                'windowWidth': 6095,
                'rescaleSlope': 1,
                'windowCenter': 23.5,
                'channelsCount': 1,
                'dimensionsIJK': {'x': 512, 'y': 512, 'z': 139}
                'IJK2WorldMatrix': [0.7617189884185793, 0, 0, -194.238403081894, 0, 0.7617189884185793, 0, -217.5384061336518, 0, 0, 2.5, -347.7500000000001, 0, 0, 0, 1],
                'rescaleIntercept': 0
            },
            path_original='/h5af-public/images/original/M/e/7R/vsytec8zX0p.nrrd',
            full_storage_url='https://dev.supervise.ly/h5un-public/images/original/M/e/7R/zX0p.nrrd',
            tags=[],
            team_id=435,
            workspace_id=685,
            project_id=18949,
            dataset_id=61803,
            file_meta={
                    'mime': 'image/nrrd',
                    'size': 46073411,
                    'type': 'int32',
                    'extra': {'stride': [1, 512, 262144],
                    'comments': []
                }
                'sizes': [512, 512, 139]
                'space': 'right-anterior-superior'
                'endian': 'little'
                'encoding': 'gzip'
                'dimension': 3
                'space origin': [-194.238403081894, -217.5384061336518, -347.7500000000001]
                'space dimension': 3
                'space directions': [[0.7617189884185793, 0, 0], [0, 0.7617189884185793, 0], [0, 0, 2.5]]
            }
            figures_count=None,
            objects_count=None,
            processing_path='1/1560071'
        )
    """

    #: :class:`int`: Volume ID in Supervisely.
    id: int

    #: :class:`str`: Volume filename.
    name: str

    #: :class:`str`: Link to volume.
    link: str

    #: :class:`str`: Volume hash obtained by base64(sha256(file_content)).
    #: Use hash for files that are expected to be stored at Supervisely or your deployed agent.
    hash: str

    #: :class:`str`: MIME type of the volume.
    mime: str

    #: :class:`str`: File extension of the volume.
    ext: str

    #: :class:`int`: Size of the volume in bytes.
    sizeb: int

    #: :class:`str`: Volume creation time. e.g. "2019-02-22T14:59:53.381Z".
    created_at: str

    #: :class:`str`: Time of last volume update. e.g. "2019-02-22T14:59:53.381Z".
    updated_at: str

    #: :class:`dict`: A dictionary containing metadata associated with the volume.
    meta: dict

    #: :class:`str`: Original path of the volume.
    path_original: str

    #: :class:`str`: Full storage URL of the volume.
    full_storage_url: str

    #: :class:`list`: Volume :class:`VolumeTag<supervisely.volume_annotation.volume_tag.VolumeTag>` list.
    #: e.g. "[{'entityId': 19581134, 'tagId': 385328, 'id': 12259702, 'value': 'some info}]"
    tags: list

    #: :class:`int`: :class:`TeamApi<supervisely.api.team_api.TeamApi>` ID in Supervisely.
    team_id: int

    #: :class:`int`: :class:`WorkspaceApi<supervisely.api.workspace_api.WorkspaceApi>` ID in Supervisely.
    workspace_id: int

    #: :class:`int`: :class:`Project<supervisely.project.project.Project>` ID in Supervisely.
    project_id: int

    #: :class:`int`: :class:`Dataset<supervisely.project.project.Dataset>` ID in Supervisely.
    dataset_id: int

    #: :class:`dict`: A dictionary containing metadata about the volume file.
    file_meta: dict

    #: :class:`int`: Number of figures in the volume.
    figures_count: int

    #: :class:`int`: Number of objects in the volume.
    objects_count: int

    #: :class:`str`: Path to the volume file on the server.
    processing_path: str


class VolumeApi(RemoveableBulkModuleApi):
    """
    API for working with :class:`Volume<supervisely.volume.volume>`. :class:`VolumeApi<VolumeApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        volume_id = 19581134
        volume_info = api.volume.get_info_by_id(volume_id) # api usage example
    """

    def __init__(self, api):
        """
        :param api: Api class object
        """
        super().__init__(api)
        self.annotation = VolumeAnnotationAPI(api)
        self.object = VolumeObjectApi(api)
        # self.frame = VideoFrameAPI(api)
        self.figure = VolumeFigureApi(api)
        self.tag = VolumeTagApi(api)

    @staticmethod
    def info_sequence():
        """
        Get list of all :class:`VolumeInfo<VolumeInfo>` field names.

        :return: List of :class:`VolumeInfo<VolumeInfo>` field names.`
        :rtype: :class:`list`
        """

        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.LINK,
            ApiField.HASH,
            ApiField.MIME,
            ApiField.EXT,
            ApiField.SIZEB3,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.META,
            ApiField.PATH_ORIGINAL,
            ApiField.FULL_STORAGE_URL,
            ApiField.TAGS,
            ApiField.TEAM_ID,
            ApiField.WORKSPACE_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.FILE_META,
            ApiField.FIGURES_COUNT,
            ApiField.ANN_OBJECTS_COUNT,
            ApiField.PROCESSING_PATH,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of :class:`VolumeInfo<VolumeInfo>` NamedTuple.

        :return: NamedTuple name.
        :rtype: :class:`str`
        """

        return "VolumeInfo"

    def _convert_json_info(self, info: dict, skip_missing=True):
        """Private method. Convert volume information from json to VolumeInfo<VolumeInfo>"""

        res = super()._convert_json_info(info, skip_missing=skip_missing)
        return VolumeInfo(**res._asdict())

    def get_list(
        self,
        dataset_id: int,
        filters=None,
        sort: Literal["id", "name", "description", "createdAt", "updatedAt"] = "id",
        sort_order: Literal["asc", "desc"] = "asc",
    ) -> List[VolumeInfo]:
        """
        Get list of information about all volumes for a given dataset ID.

        :param dataset_id: :class:`Dataset<supervisely.project.project.Dataset>` ID in Supervisely.
        :type dataset_id: int
        :param filters: List of parameters to sort output Volumes. See: https://dev.supervise.ly/api-docs/#tag/Volumes/paths/~1volumes.list/get
        :type filters: List[Dict[str, str]], optional
        :param sort: Attribute to sort the list by. The default is "id". Valid values are "id", "name", "description", "createdAt", "updatedAt".
        :type sort: :class:`str`
        :param sort_order: Order in which to sort the list. The default is "asc". Valid values are "asc" (ascending) and "desc" (descending).
        :type sort_order: :class:`str`
        :return: List of information about volumes in given dataset.
        :rtype: :class:`List[VolumeInfo]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 61803

            volume_infos = api.volume.get_list(dataset_id)
            print(volume_infos)
            # Output: [VolumeInfo(id=19581134, ...), VolumeInfo(id=19581135, ...), VolumeInfo(id=19581136, ...)]

            sorted_volume_infos = api.volume.get_list(dataset_id, sort="id", sort_order="desc")
            # Output: [VolumeInfo(id=19581136, ...), VolumeInfo(id=19581135, ...), VolumeInfo(id=19581134, ...)]

            filtered_volume_infos = api.volume.get_list(dataset_id, filters=[{'field': 'id', 'operator': '=', 'value': '19581135'}])
            print(filtered_volume_infos)
            # Output: [VolumeInfo(id=19581135, ...)]
        """

        return self.get_list_all_pages(
            "volumes.list",
            {
                ApiField.DATASET_ID: dataset_id,
                ApiField.FILTER: filters or [],
                ApiField.SORT: sort,
                ApiField.SORT_ORDER: sort_order,
            },
        )

    def get_info_by_id(self, id: int):
        """
        Get Volume information by ID in VolumeInfo<VolumeInfo> format.

        :param id: Volume ID in Supervisely.
        :type id: int
        :return: Information about Volume. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`VolumeInfo`

        :Usage example:

         .. code-block:: python

            import supervisely as sly


            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            volume_id = 19581134
            volume_info = api.volume.get_info_by_id(volume_id)
            print(volume_info)
            # Output:
            # VolumeInfo(
            #     id=19581134,
            #     name='CTChest.nrrd',
            #     link=None,
            #     hash='+0K2oFHpqA5dwRKQlhkvhXEh52cs=',
            #     mime=None,
            #     ext=None,
            #     sizeb=46073411,
            #     created_at='2023-03-29T12:30:37.078Z',
            #     updated_at='2023-03-29T12:30:37.078Z',
            #     meta={
            #         'ACS': 'RAS',
            #         'intensity': {'max': 3071, 'min': -3024},
            #         'windowWidth': 6095,
            #         'rescaleSlope': 1,
            #         'windowCenter': 23.5,
            #         'channelsCount': 1,
            #         'dimensionsIJK': {'x': 512, 'y': 512, 'z': 139},
            #         'IJK2WorldMatrix': [0.7617189884185793, 0, 0, -194.238403081894, 0, 0.7617189884185793, 0, -217.5384061336518, 0, 0, 2.5, -347.7500000000001, 0, 0, 0, 1],
            #         'rescaleIntercept': 0
            #     },
            #     path_original='/h5af-public/images/original/M/e/7R/vs0p.nrrd',
            #     full_storage_url='https://dev.supervise.ly/.../original/M/e/7R/zX0p.nrrd',
            #     tags=[],
            #     team_id=435,
            #     workspace_id=685,
            #     project_id=18949,
            #     dataset_id=61803,
            #     file_meta={
            #             'mime': 'image/nrrd',
            #             'size': 46073411,
            #             'type': 'int32',
            #             'extra': {'stride': [1, 512, 262144],
            #             'comments': []
            #         }
            #         'sizes': [512, 512, 139]
            #         'space': 'right-anterior-superior'
            #         'endian': 'little'
            #         'encoding': 'gzip'
            #         'dimension': 3
            #         'space origin': [-194.238403081894, -217.5384061336518, -347.7500000000001]
            #         'space dimension': 3
            #         'space directions': [[0.7617189884185793, 0, 0], [0, 0.7617189884185793, 0], [0, 0, 2.5]]
            #     }
            #     figures_count=None,
            #     objects_count=None,
            #     processing_path='1/1560071'
            # )
        """

        return self._get_info_by_id(id, "volumes.info")

    def upload_hash(self, dataset_id: int, name: str, hash: str, meta: dict = None):
        """
        Upload Volume from given hash to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param name: Volume name with extension.
        :type name: str
        :param hash: Volume hash.
        :type hash: str
        :param meta: A dictionary containing data associated with the volume.
        :type meta: dict
        :return: Information about Volume. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`VolumeInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dst_dataset_id = 61958
            src_volume_id = 19581134
            volume_info = api.volume.get_info_by_id(src_volume_id)
            hash = volume_info.hash
            # It is necessary to upload volume with the same extention as in src dataset
            name = volume_info.name
            meta = volume_info.meta
            new_volume_info = api.volume.upload_hash(dst_dataset_id, name, hash, meta)
            print(new_volume_info)
            # Output:
            # VolumeInfo(
            #     id=19613940,
            #     name='CTACardio.nrrd',
            #     link=None,
            #     hash='+0K2oFHpqA5dwRKQlh5bDUA0jkPsEE52cs=',
            #     mime=None,
            #     ext=None,
            #     sizeb=67614735,
            #     created_at='2023-03-29T12:30:37.078Z',
            #     updated_at='2023-03-29T12:30:37.078Z',
            #     meta={
            #         'ACS': 'RAS',
            #         'intensity': {'max': 3532, 'min': -1024},
            #         'windowWidth': 4556,
            #         'rescaleSlope': 1,
            #         'windowCenter': 1254,
            #         'channelsCount': 1,
            #         'dimensionsIJK': {'x': 512, 'y': 512, 'z': 321}
            #         'IJK2WorldMatrix': [0.9335939999999999, 0, 0, -238.53326699999997, 0, 0.9335939999999999, 0, -238.53326699999994, 0, 0, 1.25, -200.0000000000001, 0, 0, 0, 1],
            #         'rescaleIntercept': 0
            #     },
            #     path_original='/h5af-public/images/original/M/e/7R/zfsfX0p.nrrd',
            #     full_storage_url='https://dev.supervise.ly/h5un-public/images/original/M/e/7R/zXdd0p.nrrd',
            #     tags=[],
            #     team_id=435,
            #     workspace_id=685,
            #     project_id=18949,
            #     dataset_id=61958,
            #     file_meta={
            #             'mime': 'image/nrrd',
            #             'size': 46073411,
            #             'type': 'int32',
            #             'extra': {'stride': [1, 512, 262144],
            #             'comments': []
            #         }
            #         'sizes': [512, 512, 139]
            #         'space': 'right-anterior-superior'
            #         'endian': 'little'
            #         'encoding': 'gzip'
            #         'dimension': 3
            #         'space origin': [-194.238403081894, -217.5384061336518, -347.7500000000001]
            #         'space dimension': 3
            #         'space directions': [[0.7617189884185793, 0, 0], [0, 0.7617189884185793, 0], [0, 0, 2.5]]
            #     }
            #     figures_count=None,
            #     objects_count=None,
            #     processing_path='1/1560071'
            # )
        """

        metas = None if meta is None else [meta]
        return self.upload_hashes(dataset_id, [name], [hash], metas=metas)[0]

    def upload_hashes(
        self,
        dataset_id: int,
        names: List[str],
        hashes: List[str],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        metas: List[dict] = None,
    ):
        """
        Upload Volumes from given hashes to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param names: Volumes names with extension.
        :type names: List[str]
        :param hashes: Volumes hashes.
        :type hashes: List[str]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :param metas: Volumes metadata.
        :type metas: List[dict], optional
        :return: List with information about Volumes. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[VolumeInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from tqdm import tqdm

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_dataset_id = 61958
            dst_dataset_id = 55853

            hashes = []
            names = []
            metas = []
            volume_infos = api.volume.get_list(src_dataset_id)

            # Create lists of hashes, volumes names and meta information for each volume
            for volume_info in volume_infos:
                hashes.append(volume_info.hash)
                # It is necessary to upload volumes with the same names(extentions) as in src dataset
                names.append(volume_info.name)
                metas.append(volume_info.meta)

            p = tqdm(desc="api.volume.upload_hashes", total=len(hashes))
            new_volumes_info = api.volume.upload_hashes(
                dataset_id=dst_dataset_id,
                names=names,
                hashes=hashes,
                progress_cb=p,
                metas=metas,
            )

            # Output:
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Volumes upload: ", "current": 0, "total": 2, "timestamp": "2023-04-04T07:47:11.506Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Volumes upload: ", "current": 2, "total": 2, "timestamp": "2023-04-04T07:47:11.563Z", "level": "info"}
        """

        return self._upload_bulk_add(
            lambda item: (ApiField.HASH, item),
            dataset_id,
            names,
            hashes,
            progress_cb,
            metas=metas,
        )

    def _upload_bulk_add(
        self, func_item_to_kv, dataset_id, names, items, progress_cb=None, metas=None
    ):
        """
        Private method. Bulk upload volumes to Dataset.
        """

        results = []

        if len(names) == 0:
            return results
        if len(names) != len(items):
            raise RuntimeError('Can not match "names" and "items" lists, len(names) != len(items)')

        if metas is None:
            metas = [{}] * len(names)
        else:
            if len(names) != len(metas):
                raise RuntimeError('Can not match "names" and "metas" len(names) != len(metas)')

        for batch in batched(list(zip(names, items, metas))):
            volumes = []
            for name, item, meta in batch:
                item_tuple = func_item_to_kv(item)
                image_data = {ApiField.NAME: name, item_tuple[0]: item_tuple[1]}
                if len(meta) != 0 and type(meta) == dict:
                    image_data[ApiField.META] = meta
                volumes.append(image_data)

            response = self._api.post(
                "volumes.bulk.add",
                {ApiField.DATASET_ID: dataset_id, ApiField.VOLUMES: volumes},
            )
            if progress_cb is not None:
                progress_cb(len(volumes))

            for info_json in response.json():
                results.append(self._convert_json_info(info_json))
        return results

    def upload_np(
        self,
        dataset_id: int,
        name: str,
        np_data,
        meta: dict,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        batch_size: int = 30,
    ):
        """
        Upload given Volume in numpy format with given name to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param name: Volume name with extension.
        :type name: str
        :param np_data: image in RGB format(numpy matrix)
        :type np_data: np.ndarray
        :param meta: Volume metadata.
        :type meta: dict, optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about Volume. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`VolumeInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            np_volume, meta = sly.volume.read_nrrd_serie_volume_np(local_path)
            nrrd_info_np = api.volume.upload_np(
                dataset.id,
                "MRHead_np.nrrd",
                np_volume,
                meta,
            )
            print(f"Volume uploaded as NumPy array to Supervisely with ID:{nrrd_info_np.id}")

            # Output:
            # Volume uploaded as NumPy array to Supervisely with ID:18562982
        """

        ext = get_file_ext(name)
        if ext != ".nrrd":
            raise ValueError("Name has to be with .nrrd extension, for example: my_volume.nrrd")
        from timeit import default_timer as timer

        with logging_redirect_tqdm():
            logger.debug(f"Start volume {name} compression before upload...")

        start = timer()
        volume_bytes = volume.encode(np_data, meta)
        with logging_redirect_tqdm([logger]):
            logger.debug(f"Volume has been compressed in {timer() - start} seconds")
            logger.debug(f"Start uploading bytes of '{name}' volume ...")

        start = timer()
        volume_hash = get_bytes_hash(volume_bytes)
        self._api.image._upload_data_bulk(lambda v: v, [(volume_bytes, volume_hash)])

        with logging_redirect_tqdm([logger]):
            logger.debug(
                f"3D volume bytes has been sucessfully uploaded in {timer() - start} seconds"
            )

        volume_info = self.upload_hash(dataset_id, name, volume_hash, meta)
        if progress_cb is not None:
            progress_cb(1)  # upload volume

        # slice all directions
        # http://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/Python_html/03_Image_Details.html#Conversion-between-numpy-and-SimpleITK
        # x = 1 - sagittal
        # y = 1 - coronal
        # z = 1 - axial
        planes = [Plane.SAGITTAL, Plane.CORONAL, Plane.AXIAL]

        for plane, dimension in zip(planes, np_data.shape):
            for batch in batched(list(range(dimension))):
                slices = []
                slices_bytes = []
                slices_hashes = []
                try:
                    for i in batch:
                        normal = Plane.get_normal(plane)

                        if plane == Plane.SAGITTAL:
                            pixel_data = np_data[i, :, :]
                        elif plane == Plane.CORONAL:
                            pixel_data = np_data[:, i, :]
                        elif plane == Plane.AXIAL:
                            pixel_data = np_data[:, :, i]
                        else:
                            raise ValueError(f"Unknown plane {plane}")

                        img_bytes = nrrd_encoder.encode(
                            pixel_data, header={"encoding": "gzip"}, compression_level=1
                        )

                        img_hash = get_bytes_hash(img_bytes)
                        slices_bytes.append(img_bytes)
                        slices_hashes.append(img_hash)
                        slices.append(
                            {
                                "hash": img_hash,
                                "sliceIndex": i,
                                "normal": normal,
                            }
                        )

                    if len(slices) > 0:
                        self._api.image._upload_data_bulk(
                            lambda v: v,
                            zip(slices_bytes, slices_hashes),
                        )
                        self._upload_slices_bulk(volume_info.id, slices, progress_cb)

                except Exception as e:
                    exc_str = str(e)
                    logger.warn(
                        "File skipped due to error: {}".format(exc_str),
                        exc_info=True,
                        extra={
                            "exc_str": exc_str,
                            "file_path": name,
                        },
                    )
        return volume_info

    def upload_dicom_serie_paths(
        self,
        dataset_id: int,
        name: str,
        paths: List[str],
        log_progress: bool = True,
        anonymize: bool = True,
    ) -> VolumeInfo:
        """
        Upload given DICOM series from given paths to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param name: Volume name with extension.
        :type name: str
        :param paths: Local volumes paths.
        :type paths: List[str]
        :param log_progress: Determine if additional technical logs are displaying.
        :type log_progress: bool, optional
        :param anonymize: Determine whether to hide PatientID and PatientName fields.
        :type anonymize: bool, optional
        :return: Information about Volume. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`VolumeInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()


            dicom_dir_name = "src/upload/MRHead_dicom/"
            series_infos = sly.volume.inspect_dicom_series(root_dir=dicom_dir_name)

            for serie_id, files in series_infos.items():
                item_path = files[0]
                name = f"{sly.fs.get_file_name(path=item_path)}.nrrd"
                dicom_info = api.volume.upload_dicom_serie_paths(
                    dataset_id=dataset.id,
                    name=name,
                    paths=files,
                    anonymize=True,
                )
                print(f"DICOM volume has been uploaded to Supervisely with ID: {dicom_info.id}")

            # Output:
            # DICOM volume has been uploaded to Supervisely with ID: 18630608
        """

        volume_np, volume_meta = volume.read_dicom_serie_volume_np(paths, anonymize=anonymize)
        progress_cb = None
        if log_progress is True:
            progress_cb = Progress(f"Upload volume {name}", sum(volume_np.shape)).iters_done_report
        res = self.upload_np(dataset_id, name, volume_np, volume_meta, progress_cb)
        return self.get_info_by_name(dataset_id, name)

    def _upload_slices_bulk(
        self, volume_id: int, items, progress_cb: Optional[Union[tqdm, Callable]] = None
    ):
        """
        Private method for volume slices bulk uploading.

        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :param items: Volume slices to upload
        :type items: list
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: List of responses
        :rtype: list
        """

        results = []
        if len(items) == 0:
            return results

        for batch in batched(items):
            response = self._api.post(
                "volumes.slices.bulk.add",
                {ApiField.VOLUME_ID: volume_id, ApiField.VOLUME_SLICES: batch},
            )
            if progress_cb is not None:
                progress_cb(len(batch))
            results.extend(response.json())
        return results

    def upload_nrrd_serie_path(
        self, dataset_id: int, name: str, path: str, log_progress=True, progress_cb=None
    ) -> VolumeInfo:
        """
        Upload NRRD format volume from given path to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param name: Volume name with extension.
        :type name: str
        :param path: Local volume path.
        :type path: str
        :param log_progress: Determine if additional technical logs are displaying.
        :type log_progress: bool, optional
        :return: Information about Volume. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`VolumeInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()


            local_path = "src/upload/nrrd/MRHead.nrrd"

            nrrd_info = api.volume.upload_nrrd_serie_path(
                dataset.id,
                "MRHead.nrrd",
                local_path,
            )
            print(f'"{nrrd_info.name}" volume uploaded to Supervisely with ID:{nrrd_info.id}')

            # Output:
            # "NRRD_1.nrrd" volume uploaded to Supervisely with ID:18562981
        """

        volume_np, volume_meta = volume.read_nrrd_serie_volume_np(path)

        progress_nrrd = None
        if log_progress is True or progress_cb is not None:
            progress_nrrd = tqdm_sly(
                desc=f"Uploading volume '{name}'",
                total=sum(volume_np.shape),
                leave=True if progress_cb is None else False,
                position=1,
            )
        res = self.upload_np(dataset_id, name, volume_np, volume_meta, progress_nrrd)
        return self.get_info_by_name(dataset_id, name)

    def _download(self, id: int, is_stream: bool = False):
        """
        Private method for volume volume downloading.

        :param id: Volume ID in Supervisely.
        :type id: int
        :param stream: Define, if you'd like to get the raw socket response from the server.
        :type stream: bool, optional
        :return: Response object
        :rtype: :class:`Response`
        """

        response = self._api.post("volumes.download", {ApiField.ID: id}, stream=is_stream)
        return response

    def download_path(
        self, id: int, path: str, progress_cb: Optional[Union[tqdm, Callable]] = None
    ):
        """
        Download volume with given ID to local directory.

        :param id: Volume ID in Supervisely.
        :type id: int
        :param path: Local path to save volume.
        :type path: str
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about Volume. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`VolumeInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from tqdm import tqdm

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_dataset_id = 61229
            volume_infos = api.volume.get_list(src_dataset_id)
            volume_id = volume_infos[0].id
            volume_info = api.volume.get_info_by_id(id=volume_id)

            download_dir_name = "src/download/"
            path = os.path.join(download_dir_name, volume_info.name)
            if os.path.exists(path):
                os.remove(path)

            p = tqdm(desc="Volumes upload: ", total=volume_info.sizeb, is_size=True)
            api.volume.download_path(volume_info.id, path, progress_cb=p)

            if os.path.exists(path):
                print(f"Volume (ID {volume_info.id}) successfully downloaded.")

            # Output:
            # Volume (ID 18630603) successfully downloaded.
        """

        response = self._download(id, is_stream=True)
        ensure_base_path(path)

        with open(path, "wb") as fd:
            mb1 = 1024 * 1024
            for chunk in response.iter_content(chunk_size=mb1):
                fd.write(chunk)

                if progress_cb is not None:
                    progress_cb(len(chunk))

    def upload_nrrd_series_paths(
        self,
        dataset_id: int,
        names: List[str],
        paths: List[str],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        log_progress: bool = True,
    ):
        """
        Upload NRRD format volumes from given paths to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param names: Volumes names with extensions.
        :type names: List[str]
        :param paths: Local volumes paths.
        :type paths: List[str]
        :param log_progress: Determine if additional technical logs are displaying.
        :type log_progress: bool, optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about Volume. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`VolumeInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()


            local_dir_name = "src/upload/nrrd/"
            all_nrrd_names = os.listdir(local_dir_name)
            names = [f"1_{name}" for name in all_nrrd_names]
            paths = [os.path.join(local_dir_name, name) for name in all_nrrd_names]

            volume_infos = api.volume.upload_nrrd_series_paths(dataset.id, names, paths)
            print(f"All volumes has been uploaded with IDs: {[x.id for x in volume_infos]}")

            # Output:
            # All volumes has been uploaded with IDs: [18630605, 18630606, 18630607]
        """

        volume_infos = []
        for name, path in zip(names, paths):
            info = self.upload_nrrd_serie_path(dataset_id, name, path, log_progress, progress_cb)
            volume_infos.append(info)
            if progress_cb is not None:
                progress_cb(1)
        return volume_infos

    def download_slice_np(
        self,
        volume_id: int,
        slice_index: int,
        plane: Literal["sagittal", "coronal", "axial"],
        window_center: float = None,
        window_width: int = None,
    ):
        """
        Download slice as NumPy from Supervisely by ID.

        :param volume_id: Volume ID in Supervisely.
        :type volume_id: int
        :param slice_index: :py:class:`Slice<supervisely.volume_annotation.slice.Slice>` index.
        :type slice_index: int
        :param plane: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>` of the slice in volume.
        :type plane: str
        :param window_center: Window center.
        :type window_center: float
        :param window_width: Window width.
        :type window_width: int
        :return: Information about Volume. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`VolumeInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()


            slice_index = 60

            image_np = api.volume.download_slice_np(
                volume_id=volume_id,
                slice_index=slice_index,
                plane=sly.Plane.SAGITTAL,
            )

            print(f"Image downloaded as NumPy array. Image shape: {image_np.shape}")

            # Output:
            # Image downloaded as NumPy array. Image shape: (256, 256, 3)
        """

        normal = Plane.get_normal(plane)
        meta = self.get_info_by_id(volume_id).meta

        if window_center is None:
            if "windowCenter" in meta:
                window_center = meta["windowCenter"]
            else:
                window_center = meta["intensity"]["min"] + meta["windowWidth"] / 2

        if window_width is None:
            if "windowWidth" in meta:
                window_width = meta["windowWidth"]
            else:
                window_width = meta["intensity"]["max"] - meta["intensity"]["min"]

        data = {
            "volumeId": volume_id,
            "sliceIndex": slice_index,
            "normal": normal,
            "windowCenter": window_center,
            "windowWidth": window_width,
        }
        image_bytes = self._api.post(
            method="volumes.slices.images.download", data=data, stream=True
        ).content

        return read_bytes(image_bytes)

    def get_free_names(self, dataset_id: int, names: List[str]) -> List[str]:
        """
        Returns list of free names for given dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param names: List of names to check.
        :type names: List[str]
        :return: List of free names.
        :rtype: List[str]
        """

        volumes_in_dataset = self.get_list(dataset_id)
        used_names = {volume_info.name for volume_info in volumes_in_dataset}
        new_names = [
            generate_free_name(used_names, name, with_ext=True, extend_used_names=True)
            for name in names
        ]
        return new_names

    def raise_name_intersections_if_exist(
        self, dataset_id: int, names: List[str], message: str = None
    ):
        """
        Raises error if volumes with given names already exist in dataset.
        Default error message:
        "Volumes with the following names already exist in dataset [ID={dataset_id}]: {name_intersections}.
        Please, rename volumes and try again or set change_name_if_conflict=True to rename automatically on upload."
        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param names: List of names to check.
        :type names: List[str]
        :param message: Error message.
        :type message: str, optional
        :return: None
        :rtype: None
        """
        volumes_in_dataset = self.get_list(dataset_id)
        used_names = {volume_info.name for volume_info in volumes_in_dataset}
        name_intersections = used_names.intersection(set(names))
        if message is None:
            message = f"Volumes with the following names already exist in dataset [ID={dataset_id}]: {name_intersections}. Please, rename volumes and try again or set change_name_if_conflict=True to rename automatically on upload."
        if len(name_intersections) > 0:
            raise ValueError(f"{message}")

    def upload_dir(
        self,
        dataset_id: int,
        dir_path: str,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        log_progress: bool = False,
        change_name_if_conflict: bool = True,
    ) -> List[VolumeInfo]:
        """
        Upload all volumes from given directory to Dataset recursively.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param dir_path: Local directory path.
        :type dir_path: str
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :param log_progress: Determine if additional technical logs are displaying.
        :type log_progress: bool, optional
        :param change_name_if_conflict: Determine if names are changing if conflict.
        :type change_name_if_conflict: bool, optional
        :return: List with information about Volumes. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[VolumeInfo]`
        :Usage example:

             .. code-block:: python

                import supervisely as sly
                from tqdm import tqdm

                os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
                os.environ['API_TOKEN'] = 'Your Supervisely API Token'

                api = sly.Api.from_env()

                dataset_id = 61958
                dir_path = "src/upload/volumes/"
                volume_infos = api.volume.upload_dir(dataset_id, dir_path)
                print(f"All volumes has been uploaded with IDs: {[x.id for x in volume_infos]}")

                # Output:
                # All volumes has been uploaded with IDs: [18630605, 18630606, 18630607]
        """
        names = []
        paths = []

        if os.path.isdir(dir_path) is False:
            raise ValueError(f"Path {dir_path} is not a directory or does not exist")

        dcm_volumes = inspect_dicom_series(dir_path, logging=log_progress)
        for serie, files in dcm_volumes.items():
            name = f"{serie}.nrrd"
            names.append(name)
            paths.append(files)
        nrrd_volumes = inspect_nrrd_series(dir_path, logging=log_progress)
        for volume_path in nrrd_volumes:
            name = get_file_name_with_ext(volume_path)
            names.append(name)
            paths.append(volume_path)

        if change_name_if_conflict:
            names = self.get_free_names(dataset_id, names)
        else:
            self.raise_name_intersections_if_exist(dataset_id, names)

        volume_infos = []
        for name, path in zip(names, paths):
            if isinstance(path, tuple):
                volume_info = self.upload_dicom_serie_paths(
                    dataset_id=dataset_id,
                    name=name,
                    paths=path,
                    log_progress=log_progress,
                )
            else:
                volume_info = self.upload_nrrd_serie_path(
                    dataset_id=dataset_id,
                    name=name,
                    path=path,
                    log_progress=log_progress,
                )
            volume_infos.append(volume_info)
            if progress_cb is not None:
                progress_cb(1)
        return volume_infos

    def upload_dirs(
        self,
        dataset_id: int,
        dir_paths: List[str],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        log_progress: bool = False,
        change_name_if_conflict: bool = True,
    ) -> List[VolumeInfo]:
        """
        Upload all volumes from given directories to Dataset recursively.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param dir_paths: List of local directory paths.
        :type dir_paths: List[str]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :param log_progress: Determine if additional technical logs are displaying.
        :type log_progress: bool, optional
        :param change_name_if_conflict: Determine if names are changing if conflict.
        :type change_name_if_conflict: bool, optional
        :return: List with information about Volumes. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[VolumeInfo]`
        :Usage example:

             .. code-block:: python

                import supervisely as sly
                from tqdm import tqdm

                os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
                os.environ['API_TOKEN'] = 'Your Supervisely API Token'

                api = sly.Api.from_env()

                dataset_id = 61958
                dir_paths = ["src/upload/volumes1/", "src/upload/volumes2/"]
                volume_infos = api.volume.upload_dirs(dataset_id, dir_paths)
                print(f"All volumes has been uploaded with IDs: {[x.id for x in volume_infos]}")

                # Output:
                # All volumes has been uploaded with IDs: [18630605, 18630606, 18630607]
        """
        volume_infos = []
        if not isinstance(dir_paths, list):
            raise ValueError(f"dir_paths must be a list of strings, but got {type(dir_paths)}")

        for dir_path in dir_paths:
            volume_infos.extend(
                self.upload_dir(
                    dataset_id, dir_path, progress_cb, log_progress, change_name_if_conflict
                )
            )
        return volume_infos



import imghdr
import os
from typing import List

import supervisely.convert.image.sly.sly_image_helper as sly_image_helper
from supervisely import Annotation, ProjectMeta, logger
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.io.fs import JUNK_FILES, get_file_ext
from supervisely.io.json import load_json_file


class SLYImageConverter(ImageConverter):
    def __init__(self, input_data: str, labeling_interface: str) -> None:
        self._input_data: str = input_data
        self._items: List[ImageConverter.Item] = []
        self._meta: ProjectMeta = None
        self._labeling_interface = labeling_interface

    def __str__(self):
        return AvailableImageConverters.SLY

    @property
    def ann_ext(self) -> str:
        return ".json"

    @property
    def key_file_ext(self) -> str:
        return ".json"

    def generate_meta_from_annotation(self, ann_path: str, meta: ProjectMeta) -> ProjectMeta:
        meta = sly_image_helper.get_meta_from_annotation(ann_path, meta)
        return meta

    def validate_ann_file(self, ann_path: str, meta: ProjectMeta) -> bool:
        try:
            ann_json = load_json_file(ann_path)
            if "annotation" in ann_json:
                ann_json = ann_json["annotation"]
            ann = Annotation.from_json(ann_json, meta)
            return True
        except:
            return False

    def validate_key_file(self, key_file_path: str) -> bool:
        try:
            self._meta = ProjectMeta.from_json(load_json_file(key_file_path))
            return True
        except Exception:
            return False

    def validate_format(self) -> bool:
        detected_ann_cnt = 0
        meta_path = None
        images_list, ann_dict, img_meta_dict = [], {}, {}
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                dir_name = os.path.basename(root)
                if file == "meta.json":
                    is_valid = self.validate_key_file(full_path)
                    if is_valid:
                        continue

                ext = get_file_ext(full_path)
                if file in JUNK_FILES:  # add better check
                    continue
                elif ext in self.ann_ext:
                    if dir_name == "meta":
                        img_meta_dict[file] = full_path
                    else:
                        ann_dict[file] = full_path
                elif imghdr.what(full_path):
                    images_list.append(full_path)

        if self._meta is not None:
            meta = self._meta
        else:
            meta = ProjectMeta()

        # create Items
        self._items = []
        for image_path in images_list:
            item = self.Item(image_path)
            ann_name = f"{item.name}.json"
            if ann_name in ann_dict:
                ann_path = ann_dict[ann_name]
                if self._meta is None:
                    meta = self.generate_meta_from_annotation(ann_path, meta)
                is_valid = self.validate_ann_file(ann_path, meta)
                if is_valid:
                    item.ann_data = ann_path
                    detected_ann_cnt += 1
            if ann_name in img_meta_dict:
                item.set_meta_data(img_meta_dict[ann_name])
            self._items.append(item)
        self._meta = meta
        return detected_ann_cnt > 0

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        """Convert to Supervisely format."""
        if meta is None:
            meta = self._meta

        if item.ann_data is None:
            return item.create_empty_annotation()

        try:
            ann_json = load_json_file(item.ann_data)
            if "annotation" in ann_json:
                ann_json = ann_json["annotation"]
            if renamed_classes or renamed_tags:
                ann_json = sly_image_helper.rename_in_json(ann_json, renamed_classes, renamed_tags)
            return Annotation.from_json(ann_json, meta)
        except Exception as e:
            logger.warn(f"Failed to convert annotation: {repr(e)}")
            return item.create_empty_annotation()
